
# lambdas: 'vec' overwrites the sequence formed as seq(lim[1],lim[2],by = by)
# for linear kernel, set lambdas = list(vec = 1.0,...)
# weights,tt,sampling.weights can be a formula or numeric vector
cv.rts<-function(
formula,data,weights,tt,
sampling.weights = 1,
kfold = 2,
random = FALSE, # will be converted to an integer,how many times to do random cv for a given (lambda,gamma) pair
lambdas = list(vec = NULL,lim = c(.1,.5),by = .01), 
gammas  = list(vec = NULL,lim = c(.1,.5),by = .01),  # kernel parameter
beta.init = NULL,
eta.init = NULL,
cost = 0.0,threshold = 1e-4,
kernel = c("l","e","p","r","i"),
method = c("rts","svm"),
rts.control = list(maxit = 100,tol = 1e-6,verbose = 0),
optim.control = control.loqo(),
verbose = 0
){
	
	allvars <- all.vars(formula)
	if(inherits(weights,'formula')) allvars <- c(allvars,all.vars(weights))
	if(inherits(sampling.weights,'formula')) allvars <- c(allvars,all.vars(sampling.weights))
	if(inherits(tt,'formula')) allvars <- c(allvars,all.vars(tt))
	allvars <- unique(allvars)
	all.formula <- paste(allvars[1],"~",paste(allvars[-1],collapse = "+"))

	data <- model.frame(all.formula,data)
	n <- nrow(data)
	y <- model.response(data)
	if(inherits(sampling.weights,'formula'))sampling.weights <- model.matrix(update(sampling.weights,~-1+.),data)
	if(inherits(weights,'formula'))weights <- model.matrix(update(weights,~-1+.),data)
	if(inherits(tt,'formula'))tt <- model.matrix(update(tt,~-1+.),data)

	# data <- model.frame(formula,data)
	# n <- nrow(data)
	# y <- model.response(data)

	# if(!setequal(c(0,1),unique(y))) stop("Outcome 'y' should be numeric in {0,1}")
	if(substr(method[1],1,1) == "s"){ #svm
		if(!setequal(c(-1,1),unique(y))){
			warning("Outcome 'y' should be numeric in {-1,1}, trying to recode")
			y <- as.numeric(y)
			y[y == 0] <- -1
			if(!setequal(c(-1,1),unique(y)))
				stop("Could not recode outcome 'y' should to numeric in {-1,1}")
			else
				warning("Successfully recoded outcome 'y' to numeric in {-1,1}")
		}
	}
	
	# treat intercept separately. X does not contain intercept even if formula does
	X <- NULL
	INTERCEPT <- has.intercept(formula)
	formula <- update(formula,.~+1+.)
	X <- model.matrix(formula,data)[,-1,drop = FALSE]
	
	
	if(missing(weights))stop("weights must be given")
	w <- weights
    if (is.null(w) || !is.numeric(w)) 
        stop("'weights' must be a numeric vector")
	if(length(w) != n){
		warning("length of fitting weights does not match number of data points.")
		w <- rep(w,length.out = n)
	}		

	if(missing(tt))stop("tt must be given")
    if (is.null(tt) || !is.numeric(tt)) 
        stop("'tt' must be a numeric vector")
	if(length(tt) != n){
		warning("length of 'tt' does not match number of data points.")
		tt <- rep(tt,length.out = n)
	}

	if(length(sampling.weights) != n){
		warning("length of sampling weights does not match number of data points.")
		sampling.weights <- rep(sampling.weights,length.out = n)
	}	
	if(any(sampling.weights < 0)) stop("sampling weights must be nonnegative")
	
    kernel <- substr(kernel[1], 1, 1)
	is.eta.init <- !is.null(eta.init)	
	is.beta.init <- !is.null(beta.init)	
	if((kernel == "l") &&  !is.eta.init && !is.beta.init) 
		stop("Either beta.init or eta.init must be given for linear kernel")		
	kernel <- match(kernel,c("l","e","p","r","i"),nomatch = 0) - 1
	if(kernel < 0) stop("Invalid kernel selected.")

	nlambda <- length(lambdas$vec)
	if(nlambda == 0) lambdas$vec <- double(1) # we cannot pass NULL via .C(), so nlambda == 0 indicates C to use from/by/to fields
	if(!is.finite(nlambda)) stop("error in generating lambda parameter sequence")
	NLAMBDA <- nlambda
	if(NLAMBDA == 0)NLAMBDA <- as.integer(ceiling(diff(lambdas$lim) / lambdas$by) + 1)
	
	ngamma <- length(gammas$vec)
	if(ngamma == 0) gammas$vec <- double(1)
	if(!is.finite(ngamma)) stop("error in generating kernel parameter sequence")
	NGAMMA <- ngamma
	if(NGAMMA == 0)NGAMMA <- as.integer(ceiling(diff(gammas$lim) / gammas$by) + 1)
	
	# LOQO control variables are passed to C all as 'double'-s 
	optim.control <- control.loqo(control = optim.control)
	optim.control$solver <- as.integer(match(tolower(optim.control$solver),c("ldl","lu","qr"),nomatch = 1)-1)
	optim.control <- .as.double(unlist(optim.control))
 
	rts.control <- control.rts(control = rts.control)
 
	y <- .as.double(y)
	X <- .as.double(X)
	w <- .as.double(w) 
	tt <- .as.double(tt)  	

	random <- as.integer(max(0,random));
	dim.result <- c(NLAMBDA * NGAMMA,if(random==0)(kfold + 2) else (random + 2))
	res <- .C("R_cv_rts",
		as.integer(kfold),
		random,
		nlambda = nlambda,lambda=as.double(lambdas$vec), 
		from.lambda = as.double(lambdas$lim[1]),to.lambda = as.double(lambdas$lim[2]),by.lambda = as.double(lambdas$by),
		ngamma = ngamma,gamma=as.double(gammas$vec), 
		from.gamma = as.double(gammas$lim[1]),to.gamma = as.double(gammas$lim[2]),by.gamma = as.double(gammas$by),
		as.integer(INTERCEPT),as.double(threshold),as.double(cost),
		is.beta.init = as.integer(is.beta.init),beta.init = as.double(beta.init),
		as.integer(is.eta.init),as.double(eta.init),
		rts.method = as.integer(match(substr(tolower(method[1]),1,1),c("r","s"),nomatch = 1)-1), # defaults to '0' = 'rts', 1='svm'
		rts.control = as.double(rts.control[order(names(rts.control))]),
		optim.control = optim.control,
		verbose = as.integer(verbose),
		kernel.type = as.integer(kernel),
		ncx = ncol(X),nrx = nrow(X),X = X,y = y,w = w,tt = tt,sampling.weights = as.double(sampling.weights),
		result = double(prod(dim.result)),
		DUP = FALSE)
	result <- res$result
	dim(result) <- dim.result
	colnames(result) <- c("lambda","gamma",paste("theta",1:(dim.result[2]-2)))
	result
}

# weights,tt,sampling.weights can be a formula or numeric vector
R.cv.rts<-function(
formula,data,weights,tt,
sampling.weights = 1,
kfold = 2,
random = FALSE, # will be converted to an integer,how many times to do random cv for a given (lambda,gamma) pair
lambdas = list(vec = NULL,lim = c(.1,.5),by = .1),  
gammas  = list(vec = NULL,lim = c(.1,.5),by = .1),  # kernel parameter
beta.init = NULL,
eta.init = NULL,
cost = 0.0,threshold = 1e-4,
kernel = c("l","e","p","r","i"),
method = c("rts","svm"),
rts.control = list(maxit = 100,tol = 1e-6,verbose = 0),
optim.control = control.loqo(),
verbose = 0
){


	allvars <- all.vars(formula)
	if(inherits(weights,'formula')) allvars <- c(allvars,all.vars(weights))
	if(inherits(sampling.weights,'formula')) allvars <- c(allvars,all.vars(sampling.weights))
	if(inherits(tt,'formula')) allvars <- c(allvars,all.vars(tt))
	allvars <- unique(allvars)
	all.formula <- paste(allvars[1],"~",paste(allvars[-1],collapse = "+"))

	data <- model.frame(all.formula,data)
	n <- nrow(data)
	y <- model.response(data)
	if(inherits(sampling.weights,'formula'))sampling.weights <- model.matrix(update(sampling.weights,~-1+.),data)
	if(inherits(weights,'formula'))weights <- model.matrix(update(weights,~-1+.),data)
	if(inherits(tt,'formula'))tt <- model.matrix(update(tt,~-1+.),data)
	
	# data <- model.frame(formula,data)
	# n <- nrow(data)
	# y <- model.response(data)
	# # X <- model.matrix(update(formula,~-1+.),data)
	
	if(missing(weights))stop("weights must be given")
	w <- weights
    if (is.null(w) || !is.numeric(w)) 
        stop("'weights' must be a numeric vector")
	if(length(w) != n){
		warning("length of fitting weights does not match number of data points.")
		w <- rep(w,length.out = n)
	}		

	if(missing(tt))stop("tt must be given")
    if (is.null(tt) || !is.numeric(tt)) 
        stop("'tt' must be a numeric vector")
	if(length(tt) != n){
		warning("length of 'tt' does not match number of data points.")
		tt <- rep(tt,length.out = n)
	}

	if(length(sampling.weights) != n){
		warning("length of sampling weights does not match number of data points.")
		sampling.weights <- rep(sampling.weights,length.out = n)
	}
	if(any(sampling.weights < 0)) stop("sampling weights must be nonnegative")

	is.eta.init <- !is.null(eta.init)	
	is.beta.init <- !is.null(beta.init)	
	if((substr(kernel,1,1) == "l") &&  !is.eta.init && !is.beta.init) 
		stop("Either beta.init or eta.init must be given for linear kernel")		

	nlambda <- length(lambdas$vec)
	if(nlambda == 0) nlambda <- as.integer(ceiling(diff(lambdas$lim) / lambdas$by) + 1)
	if(!is.finite(nlambda)) stop("error in generating lambda parameter sequence")
	if(is.null(lambdas$vec)) lambdas$vec <- seq(from = lambdas$lim[1],to=lambdas$lim[2],by = lambdas$by)
	
	ngamma <- length(gammas$vec)
	if(ngamma == 0) ngamma <- as.integer(ceiling(diff(gammas$lim) / gammas$by) + 1)
	if(!is.finite(ngamma)) stop("error in generating kernel parameter sequence")
	if(is.null(gammas$vec)) gammas$vec <- seq(from = lambdas$lim[1],to=gammas$lim[2],by = gammas$by)

	rts.control <- control.rts(control = rts.control)
	
	# if(method[1] != "rts") stop("svm is not yet implemented")
	fitfun <- if(substr(method[1],1,1) == "r") rts.w1 else rtswt.svm #kernlab.svm
	predfun <-  predict.rtw2 #if(method[1] == "rts") predict.rtw2 else predict.svml2
	
	
	random <- as.numeric(random)
	random <- as.integer(max(0,random))
	result <- NULL
	
	if(random == 0){
	
		result <- matrix(NA,nlambda * ngamma,kfold + 2)
		colnames(result) <- c("lambda","gamma",paste("theta",1:kfold))
		
		whole <- n %/% kfold
		fold.sizes <- rep(whole,kfold)
		fold.sizes[kfold] <- fold.sizes[kfold] + n %% kfold 
		
		INDEX <- sample.int(n,size = n,replace = FALSE)
		index <- INDEX
	
		for(k in 1:kfold){
			
			kx <- 1:fold.sizes[k]
			ix <- index[kx]
			nix <- setdiff(INDEX,ix)
			
			data.fit <- data[ix,,drop = FALSE]
			w.fit <- w[ix]
			# if given, this needs to be indexed in cross-validation
			eta.init.fit <- if(is.null(eta.init)) NULL else eta.init[ix]
			data.pred <- data[nix,,drop = FALSE]
			w.sampling <- sampling.weights[nix]
			tt.pred<- tt[nix]
			y.pred <- y[nix]
			
			rx <- 1 # row index to 'result'
			for(j in 1:ngamma){
				gamma <- gammas$vec[j]
				for(i in 1:nlambda){
					lambda <- lambdas$vec[i] * nrow(data.fit)
				
					# fit
					fit <- fitfun(formula,data.fit,weights = w.fit,
						beta.init = beta.init, eta.init = eta.init.fit,
						s = 1, lambda = lambda,threshold = threshold,cost=cost,  
						kernel = kernel[1],para = gamma,
						maxit = rts.control$maxit,tol = rts.control$tol,verbose = rts.control$verbose, 
						optim.control = optim.control)
					
					rx <- (j-1) * nlambda + i
					result[rx,1] <- lambdas$vec[i]
					result[rx,2] <- gamma
					
					if(!fit$converged) next
					# predict
					pred<-predfun(fit,newdata=data.pred, case.percentage = NULL)
					theta <- caltheta.weights(y.pred,tt.pred,pred,w.sampling,cost)
					
					result[rx,k + 2] <- theta
					
					if(verbose > 0){
						if(substr(kernel[1],1,1) == "l")cat(coef(fit),"\n")
						cat(result[rx,],"\n")
						cat("\n")
					}
				}
			}	
			index <- index[-kx]
		}
	}else{
		result <- matrix(NA,nlambda * ngamma,random + 2)
		colnames(result) <- c("lambda","gamma",paste("theta",1:random))
		
		n.predict <- n %/% kfold # size to fit
		n.fit   <- n - n.predict   # size to predict
		kx    <- 1:n.fit       # n=100,kfold=5,kx=1:80

		for(j in 1:ngamma){
			gamma <- gammas$vec[j]
			for(i in 1:nlambda){
				lambda <- lambdas$vec[i] * n.fit

				rx <- (j-1) * nlambda + i # row index to 'result
				result[rx,1] <- lambdas$vec[i]
				result[rx,2] <- gamma
					
				for(r in 1:random){

					index <- sample.int(n,size = n,replace = FALSE)
					ix <- index[kx]
					nix <- setdiff(index,ix)
					
					data.fit <- data[ix,,drop = FALSE]
					w.fit <- w[ix]
					# if given, this needs to be indexed in cross-validation
					eta.init.fit <- if(is.null(eta.init)) NULL else eta.init[ix]
					data.pred <- data[nix,,drop = FALSE]
					w.sampling <- sampling.weights[nix]
					tt.pred<- tt[nix]
					y.pred <- y[nix]
			
					# fit
					fit <- fitfun(formula,data.fit,weights = w.fit,
						beta.init = beta.init, eta.init = eta.init.fit,
						s = 1, lambda = lambda,threshold = threshold,cost=cost,  
						kernel = kernel[1],para = gamma,
						maxit = rts.control$maxit,tol = rts.control$tol,verbose = rts.control$verbose, 
						optim.control = optim.control)
					
					if(!fit$converged) next
					# predict
					pred<-predfun(fit,newdata=data.pred, case.percentage = NULL)
					theta <- caltheta.weights(y.pred,tt.pred,pred,w.sampling,cost)
					
				
					if(!is.na(theta) && (is.nan(theta) || (theta == Inf))){
						warning("theta = Inf or NaN occured because sum(tt)=0 or sum(1-tt)=0")
						# print(lambda);print(gamma)
						# return(list(fit=fit,index = index,ix = ix,nix = nix,
									# data = data,w = w,tt = tt,y=y,w.sampling = w.sampling)
						# )
					}
					result[rx,r + 2] <- theta
					
					if(verbose > 0){
						if(substr(kernel[1],1,1) == "l")cat(coef(fit),"\n")
						cat(result[rx,],"\n")
						cat("\n")
					}					
				}#r
			}# i
		}# j
	}
	result
}

