#include "rts.h"
#include "random.h"
#include "matrix.h"
#include <stdlib.h>
#include <string.h>
#include <R.h>

#define	MAX(A,B)	((A) > (B) ? (A) : (B))
#define	MIN(A,B)	((A) < (B) ? (A) : (B))
#define	ABS(x)  	((x) > 0 ? (x) : (-(x)))
#define	SIGN(x)	    ((x) > 0 ? 1 : (x)==0 ? 0 : -1) 
#define EQ(x,y,eps) ((ABS((x)-(y))) <= (eps) ? 1 : 0) 
#define CX(i,j,ld)  (((j)*(ld))+(i)) //column-major index
#define RX(i,j,ld)  (((i)*(ld))+(j)) //row-major index

void (*fitfun)(RTS_MODEL_FRAME*,RTS_PARAM*,KERNEL_PARAM*,RTS_FIT* arg);
void (*predictfun)(RTS_FIT* fit,int nrx,double* x,double* result);

// return matrix _{nparam x nfold} of 'theta' values computed by rts_caltheta_weights
// we use extra memory instead of a more complicated in-situ algorithm (in memory)

// not in-situ algorithm 
// cross-validation could be done 'IN-PLACE,in-situ', i.e. without having to allocate extra memory, by
// (i) reorder the dataset so that always the first 1:k rows are used for fitting while the rest for predicting/testing
// (iia) due to requirement of contiguous memory in LAPACK matrix routines, if a subset-matrix is to be in contiguous
// memory, it'd have to be stored in ROW_MAJOR order if the matrix multipl. routine is in COLUMN_MAJOR order and vice versa
// this could be done with CLAPACKe where one can choose the storage-order of the matrix calling lin. algebra routines
// (iib) or alternatively, one might be able to use the 'LDA', leading dimension of the matrix argument in LAPACK routines



void cv_rts(
int kfold,
int random,
double cost,
double* sampling_weights,
RTS_PARAM_SEQ* lambda_seq,
RTS_PARAM_SEQ* gamma_seq, // kernel parameter seq
RTS_METHOD method,
KERNEL_PARAM* kernel,
RTS_PARAM* param, 
RTS_MODEL_FRAME* mf,
int verbose,
double* result // _{ (lambda_seq->nparam * gamma_seq->nparam) X (kfold + 2)}
){
	RTS_FIT Chunk_fit;RTS_FIT* chunk_fit = &Chunk_fit;
	RTS_MODEL_FRAME Chunk_k_mf;RTS_MODEL_FRAME* chunk_k_mf = &Chunk_k_mf;
	RTS_MODEL_FRAME Chunk_nk_mf;RTS_MODEL_FRAME* chunk_nk_mf = &Chunk_nk_mf;
	RTS_PARAM Chunk_param;RTS_PARAM* chunk_param = &Chunk_param;
	int i,j,k;
	int n = mf->nrx;
	int p = mf->ncx;
	double** ddmf = NULL;
	double* dmf_k_fold = NULL;
	double* dmf_nk_fold = NULL;
	int* index = (int*)calloc(n , sizeof(int));	
	int* perm = (int*)calloc(n , sizeof(int));	
	int dim_result[2];
	int fold_sizes[kfold];
	int whole = n / kfold;	
	int max_chunksize_k,max_chunksize_nk;
	// ncol = number of all variables involved in cross-validation: y,X,w,tt,sampling_weights,eta_init
	int ncol = (1 + mf->ncx + 1 + 1 + 1);
	if(param->eta_init) ncol++; // add into cross validation


	if(random == 0){
		for(i = 0;i < kfold - 1;i++) fold_sizes[i] = whole;
		fold_sizes[kfold - 1] = n - (kfold - 1) * whole;	
		max_chunksize_k = MAX(whole,fold_sizes[kfold - 1]);	
		max_chunksize_nk = n - MIN(whole,fold_sizes[kfold - 1]);	
	}else{
		max_chunksize_k = n - whole;
		max_chunksize_nk = n - max_chunksize_k;		
	}
	if(verbose > 0){
		Rprintf("number of folds=%i dim(%i,%i) number of variables in CV(y,X,W,tt,sampling weights, eta_init) = %i max_chunk_k(%i) max_chunk_nk(%i)\n",kfold,n,p,ncol,max_chunksize_k,max_chunksize_nk);
		if(random == 0){
			Rprintf("fold sizes=");for(i = 0;i < kfold;i++) Rprintf("%i ", fold_sizes[i]);Rprintf("\n");
		}
		Rprintf("\n");
	}
	
	// allocate space for output of cross validation
	memset(chunk_k_mf,0,sizeof(RTS_MODEL_FRAME)); 
	memset(chunk_fit,0,sizeof(RTS_FIT));
	chunk_fit->beta = (double*)calloc(mf->ncx + 1 , sizeof(double));
	chunk_fit->eta = (double*)calloc(max_chunksize_k , sizeof(double));
	chunk_fit->alpha_pred = (double*)calloc(max_chunksize_k , sizeof(double));
	chunk_fit->linear_combination = (double*)calloc(MAX(max_chunksize_k,max_chunksize_nk) , sizeof(double)); // use in both fitting and predicting

	// dmf_k_fold contains a chunk of model frame and 'eta_init' in contiguous memory 	
	dmf_k_fold = (double*)calloc(ncol * max_chunksize_k , sizeof(double));
	dmf_nk_fold = (double*)calloc(ncol * max_chunksize_nk , sizeof(double));

	fitfun = rtswt;
	predictfun = predict_rtsw;
	if(method == RTS_SVM){
		fitfun = rtswt_svm;
		predictfun = predict_svmw;
	}	
	
// check rts param	
	if(param){
		memcpy(chunk_param,param,sizeof(RTS_PARAM));
		// aliases to eta_init and beta_init are also set
	}else{
		memset(chunk_param,0,sizeof(RTS_PARAM));
		init_param_rts(chunk_param);
	}
	if(verbose > 1){
		Rprintf("Chunk parameter:\n");print_param_rts(chunk_param);Rprintf("\n");
	}
	// ddmf points to all columns in 'mf' and 'eta_init'
	ddmf = (double**)calloc(ncol , sizeof(double*));
	for(j = 0;j < p;j++) ddmf[j] = mf->X + j*mf->nrx;
	ddmf[j++] = mf->y;
	ddmf[j++] = mf->w;
	ddmf[j++] = mf->tt;
	if(param->eta_init)	ddmf[j++] = param->eta_init;
	ddmf[j++] = sampling_weights;
	if(j != ncol){
		Rprintf("LOGIC ERROR, returning\n");
		free(ddmf);
	}
	// Rprintf("ddmf\n");
	// for(j = 0;j < ncol;j++){
		// Rprintf("column %i=",j);
		// for(i = 0;i < n;i++){
			// Rprintf("%.3e ",ddmf[j][i]);
		// }
		// Rprintf("\n");
	// }
	// Rprintf("\n");	
	
	
	
	if(random == 0){

/// NON-RANDOM EXHAUSTIVE(by group) CROSS VALIDATION
/// for n = 100, kfold = 5, do 5 times the following: permute data, fit on 20 and predict on 80

		// leading(row) dimension for column-major indexing of 'result'
		dim_result[0] = lambda_seq->nparam * gamma_seq->nparam;
		dim_result[1] = kfold + 2;	

		// create a random permutation of rows stored in index
		SampleNoReplace(n,n,index,perm); // from {1..n}
		for(i = 0;i < n;i++)index[i]--; // make it 0-based	
		if(verbose > 3){
			Rprintf("permutation index=");for(i = 0;i < n;i++) Rprintf("%i ", index[i]);Rprintf("\n");
		}
			
		int* is_chunk_k = perm; //indicates if 'ith' row belongs to 'kth-chunk'
		int* kx = index; // 'kx' moves along index by 'chunksize'
		for(k = 0;k < kfold;k++){
			int chunksize_k = fold_sizes[k];
			int chunksize_nk = n - chunksize_k;
			if(verbose > 1)
				Rprintf("fitting chunk(%i) of chunksize_k(%i); predicting at chunksize_nk(%i); index pointer (%i) \n",k+1,chunksize_k,chunksize_nk,kx - index);
			memset(is_chunk_k,0,n * sizeof(int));
			for(i = 0;i < chunksize_k;i++)is_chunk_k[kx[i]] = 1; 
			// partition data as k-th fold chunk and the remainder 
			int i_k = 0;
			int i_nk = 0;
			for(i = 0;i < n;i++){
				int ii = index[i];
				// if(verbose > 3)Rprintf("processing: index[%i] = %i is_chunk(%i) i_k(%i) i_nk(%i) \n",i,ii,is_chunk_k[ii],i_k,i_nk);
				if(is_chunk_k[ii]){
					for(j = 0;j < ncol;j++)	{
						if(verbose > 4) Rprintf("ddmf[%i,%i] = %.3e \n",ii,j,ddmf[j][ii]);
						dmf_k_fold[CX(i_k,j,chunksize_k)] = ddmf[j][ii];
					}
					i_k++;
				}else{
					for(j = 0;j < ncol;j++){	
						if(verbose > 4)Rprintf("ddmf[%i,%i] = %.3e \n",ii,j,ddmf[j][ii]);
						dmf_nk_fold[CX(i_nk,j,chunksize_nk)] = ddmf[j][ii];
					}
					i_nk++;
				}
			}		
			// Rprintf("max range(mf_k)=%i range(mf_nk)=%i \n",ncol * max_chunksize,ncol * (n - min_chunksize));
			// Rprintf("cur range(mf_k)=%i range(mf_nk)=%i \n",CX(i_k,j-1,chunksize_k),CX(i_nk,j-1,chunksize_nk));		
			// Rprintf("i_k(%i) i_nk(%i)\n",i_k,i_nk);
			
			kx += chunksize_k;
			
			// must set each time (due to contiguous memory) since 'chunksize_k' is not necessarily the same
			// fit
			double* d = dmf_k_fold;
			chunk_k_mf->X = d;d+=chunksize_k * p;
			chunk_k_mf->y = d;d+=chunksize_k;
			chunk_k_mf->w = d;d+=chunksize_k;
			chunk_k_mf->tt = d;d+=chunksize_k;
			if(param->eta_init)chunk_param->eta_init = d;d+=chunksize_k;	
			chunk_k_mf->nrx = chunksize_k;
			chunk_k_mf->ncx = p;

			if(verbose > 2){
				int print_X = 0;if(verbose > 3)print_X = 1;
				Rprintf("Chunk_k(%i):\n",k+1);print_mf_rts(chunk_k_mf,print_X);Rprintf("\n");
			}
			
			// predict
			d = dmf_nk_fold;		
			chunk_nk_mf->X = d;d+=chunksize_nk * p;
			chunk_nk_mf->y = d;d+=chunksize_nk;
			chunk_nk_mf->w = d;d+=chunksize_nk;		
			chunk_nk_mf->tt = d;d+=chunksize_nk;		
			double* sampling_weights_nk = d;d+=chunksize_nk;
			double* lin_comb = chunk_fit->linear_combination;
			chunk_nk_mf->nrx = chunksize_nk;
			chunk_nk_mf->ncx = p;
			
			if(verbose > 2){
				int print_X = 0;if(verbose > 3)print_X = 1;
				Rprintf("Chunk_nk(%i):\n",k+1);print_mf_rts(chunk_nk_mf,print_X);Rprintf("\n");
			}

			for(int _j = 0;_j < gamma_seq->nparam;_j++){
				double gamma = 0.0;
				if(gamma_seq->param){
					gamma = gamma_seq->param[_j];
				}else{
					gamma = gamma_seq->from + (double)(_j) * gamma_seq->by;
					gamma = MIN(gamma,gamma_seq->to);// could happen only when j = nparam - 1
				}
				for(int _i = 0;_i < lambda_seq->nparam;_i++){
					double lambda = 0.0;
					if(lambda_seq->param){
						lambda = lambda_seq->param[_i];
					}else{
						lambda = lambda_seq->from + (double)(_i) * lambda_seq->by;
						lambda = MIN(lambda,lambda_seq->to);// could happen only when i = nparam - 1
					}
					
					int rx = _j*lambda_seq->nparam + _i; // row index to 'result'	
					//result = (lambda,gamma,theta_1 .. theta_kfold) 			
					result[CX(rx,1,dim_result[0])] = gamma;
					result[CX(rx,0,dim_result[0])] = lambda;
					result[CX(rx,k + 2,dim_result[0])] = NAN;

					// fit and predict at (lambda,gamma)			
					chunk_param->lambda = lambda * (double)chunksize_k;
					kernel->param = gamma;			
								
					fitfun(chunk_k_mf,chunk_param,kernel,chunk_fit);
					if(verbose > 1){
						if(kernel->type == LINEAR){
							Rprintf("beta=");for(j = 0;j < p + 1;j++)Rprintf("%.3e ",chunk_fit->beta[j]);Rprintf("\n");
						}
					}
					if(chunk_fit->convergence) continue;
					
					predictfun(chunk_fit,chunksize_nk,chunk_nk_mf->X,lin_comb);
					double theta = 0.0;
					
					rts_caltheta_weights(&chunksize_nk,chunk_nk_mf->y,sampling_weights_nk,chunk_nk_mf->tt,lin_comb,&cost,&theta);		
					result[CX(rx,k + 2,dim_result[0])] = theta;
					if(verbose > 0)
						Rprintf("fold(%i) lambda(%.4e) gamma(%.4e) theta(%.4e)\n\n",k+1,lambda,gamma,theta);		
				}//i
			}//j
		}//k
	}else{
/// RANDOM CROSS VALIDATION
/// for n = 100, kfold = 5, random = 18, do 18 times the following: permute data, fit on 80 and predict on 20

		// result = (lambda,gamma,theta_1 .. theta_R) R = random
		// leading(row) dimension for column-major indexing of 'result'
		dim_result[0] = lambda_seq->nparam * gamma_seq->nparam;
		dim_result[1] = random + 2;
	
		for(int _j = 0;_j < gamma_seq->nparam;_j++){
			double gamma = 0.0;
			if(gamma_seq->param){
				gamma = gamma_seq->param[_j];
			}else{
				gamma = gamma_seq->from + (double)(_j) * gamma_seq->by;
				gamma = MIN(gamma,gamma_seq->to);// could happen only when j = nparam - 1
			}
			for(int _i = 0;_i < lambda_seq->nparam;_i++){
				double lambda = 0.0;
				if(lambda_seq->param){
					lambda = lambda_seq->param[_i];
				}else{
					lambda = lambda_seq->from + (double)(_i) * lambda_seq->by;
					lambda = MIN(lambda,lambda_seq->to);// could happen only when i = nparam - 1
				}

				int rx = _j*lambda_seq->nparam + _i; // row index to 'result'	
				result[CX(rx,0,dim_result[0])] = lambda; 
				result[CX(rx,1,dim_result[0])] = gamma;
							
				for(k = 0;k < random;k++){
		
				// create a random permutation of rows stored in index	
					SampleNoReplace(n,n,index,perm); // from {1..n}
					for(i = 0;i < n;i++)index[i]--; // make it 0-based	
					if(verbose > 3){
						Rprintf("permutation index=");for(i = 0;i < n;i++) Rprintf("%i ", index[i]);Rprintf("\n");
					}
				
					int* is_chunk_k = perm; //indicates if 'ith' row belongs to 'kth-chunk'
					int* kx = index; // 'kx' moves along index by 'chunksize'

					int chunksize_k = max_chunksize_k; // i.e. = 80 for n = 100, kfold = 5, as opposed to 20 in non-random CV
					int chunksize_nk = n - chunksize_k;// i.e. = 20 for n = 100, kfold = 5, as opposed to 80 in non-random CV
					if(verbose > 1)
						Rprintf("fitting chunk(%i) of chunksize_k(%i); predicting at chunksize_nk(%i); index pointer (%i) \n",k+1,chunksize_k,chunksize_nk,kx - index);
					memset(is_chunk_k,0,n * sizeof(int));
					for(i = 0;i < chunksize_k;i++)is_chunk_k[kx[i]] = 1; 
				
					// partition data as k-th fold chunk and the remainder 
					int i_k = 0;
					int i_nk = 0;
					for(int i = 0;i < n;i++){
						int ii = index[i];
						// if(verbose > 3)Rprintf("processing: index[%i] = %i is_chunk(%i) i_k(%i) i_nk(%i) \n",i,ii,is_chunk_k[ii],i_k,i_nk);
						if(is_chunk_k[ii]){
							for(int j = 0;j < ncol;j++)	{
								if(verbose > 4) Rprintf("ddmf[%i,%i] = %.3e \n",ii,j,ddmf[j][ii]);
								dmf_k_fold[CX(i_k,j,chunksize_k)] = ddmf[j][ii];
							}
							i_k++;
						}else{
							for(int j = 0;j < ncol;j++){	
								if(verbose > 4)Rprintf("ddmf[%i,%i] = %.3e \n",ii,j,ddmf[j][ii]);
								dmf_nk_fold[CX(i_nk,j,chunksize_nk)] = ddmf[j][ii];
							}
							i_nk++;
						}
					}		
					// Rprintf("max range(mf_k)=%i range(mf_nk)=%i \n",ncol * max_chunksize,ncol * (n - min_chunksize));
					// Rprintf("cur range(mf_k)=%i range(mf_nk)=%i \n",CX(i_k,j-1,chunksize_k),CX(i_nk,j-1,chunksize_nk));		
					// Rprintf("i_k(%i) i_nk(%i)\n",i_k,i_nk);
				
					
					// must set each time (due to contiguous memory) since 'chunksize_k' is not necessarily the same
					// fit
					double* d = dmf_k_fold;
					chunk_k_mf->X = d;d+=chunksize_k * p;
					chunk_k_mf->y = d;d+=chunksize_k;
					chunk_k_mf->w = d;d+=chunksize_k;
					chunk_k_mf->tt = d;d+=chunksize_k;
					if(param->eta_init)chunk_param->eta_init = d;d+=chunksize_k;	
					chunk_k_mf->nrx = chunksize_k;
					chunk_k_mf->ncx = p;

					if(verbose > 2){
						int print_X = 0;if(verbose > 3)print_X = 1;
						Rprintf("Chunk_k(%i):\n",k+1);print_mf_rts(chunk_k_mf,print_X);Rprintf("\n");
					}
				
					// predict
					d = dmf_nk_fold;		 
					chunk_nk_mf->X = d;d+=chunksize_nk * p;
					chunk_nk_mf->y = d;d+=chunksize_nk;
					chunk_nk_mf->w = d;d+=chunksize_nk;		
					chunk_nk_mf->tt = d;d+=chunksize_nk;		
					double* sampling_weights_nk = d;d+=chunksize_nk;
					double* lin_comb = chunk_fit->linear_combination;
					chunk_nk_mf->nrx = chunksize_nk;
					chunk_nk_mf->ncx = p;
					
					if(verbose > 2){
						int print_X = 0;if(verbose > 3)print_X = 1;
						Rprintf("Chunk_nk(%i):\n",k+1);print_mf_rts(chunk_nk_mf,print_X);Rprintf("\n");
					}

					result[CX(rx,k + 2,dim_result[0])] = NAN;

					// fit and predict at (lambda,gamma)			
					chunk_param->lambda = lambda * (double)chunksize_k;
					kernel->param = gamma;			
								
					fitfun(chunk_k_mf,chunk_param,kernel,chunk_fit);
					if(verbose > 1){
						if(kernel->type == LINEAR){
							Rprintf("beta=");for(j = 0;j < p + 1;j++)Rprintf("%.3e ",chunk_fit->beta[j]);Rprintf("\n");
						}
					}
					if(chunk_fit->convergence) continue;
					
					predictfun(chunk_fit,chunksize_nk,chunk_nk_mf->X,lin_comb);
					if(verbose > 2){
						Rprintf("prediction: ");print_matrix(1,chunksize_nk,lin_comb);Rprintf("\n");
					}					
					
					double theta = 0.0;
					rts_caltheta_weights(&chunksize_nk,chunk_nk_mf->y,sampling_weights_nk,chunk_nk_mf->tt,lin_comb,&cost,&theta);		
					result[CX(rx,k + 2,dim_result[0])] = theta;
					if(verbose > 0)
						Rprintf("fold(%i) lambda(%.4e) gamma(%.4e) theta(%.4e)\n\n",k+1,lambda,gamma,theta);		
				}//r
			}//i
		}//j	
	}	
	
	
	free(chunk_fit->beta);
	free(chunk_fit->eta);
	free(chunk_fit->alpha_pred);
	free(chunk_fit->linear_combination);
	free(ddmf);
	free(dmf_k_fold);
	free(dmf_nk_fold);
	free(index);free(perm);
}

void R_cv_rts(
int* kfold,
int* random,
// penalty parameter lambda
int* nlambda,double* lambdas, 
double* from_lambda,double* to_lambda,double* by_lambda,

// kernel parameter
int* ngamma,double* gammas, 
double* from_gamma,double* to_gamma,double* by_gamma,

int* is_intercept,
double* threshold,double* cost,
int* is_beta_init,double* beta_init,
int* is_eta_init,double* eta_init,
int* rts_method,
double* rts_control,// maxit,tol,verbose,
double* loqo_control,
int* verbose,
int* kernel,
int* ncx,int* nrx,double* X,double* y,double* w,double* tt,double* sampling_weights,
double* result //_{nlambda x (kfold + 1)}
)
{
	RTS_PARAM_SEQ Lambda_seq;
	RTS_PARAM_SEQ Gamma_seq;
	KERNEL_PARAM Kernel;
	RTS_PARAM Param; 
	RTS_MODEL_FRAME mf;
	
	memset(&mf,0,sizeof(RTS_MODEL_FRAME));
	mf.y = y;
	mf.X = X;
	mf.w = w;
	mf.tt = tt;
	mf.nrx = *nrx;
	mf.ncx = *ncx;

	// Rprintf("dim(%i %i)\n",mf.nrx,mf.ncx);
	// Rprintf("R: Cross Validation Data:\n");print_mf_rts(&mf);
	
	memset(&Param,0,sizeof(RTS_PARAM));
	Param.is_intercept = *is_intercept;
	Param.s = 1.0;
	Param.threshold = *threshold;
	// Param.cost = *cost;
	if(*is_beta_init){
		Param.nbeta = *ncx + 1;
		Param.beta_init = beta_init;
	}
	if(*is_eta_init){
		Param.neta = *nrx;
		Param.eta_init = eta_init;
	}
	Param.control.maxit = rts_control[0];
	Param.control.tol = rts_control[1];
	Param.control.verbose = rts_control[2];
	Param.loqo_control.bound = loqo_control[0]; 
	Param.loqo_control.inf = loqo_control[1];
	Param.loqo_control.margin = loqo_control[2]; 
	Param.loqo_control.maxiter = (int)loqo_control[3]; 
	Param.loqo_control.sigfig_max = loqo_control[4];
	Param.loqo_control.lapack_solver = (LAPACK_SOLVER)loqo_control[5];
	Param.loqo_control.verbose = (int)loqo_control[6];		
	
	
	memset(&Kernel,0,sizeof(KERNEL_PARAM));
	Kernel.type = (KERNEL_TYPE)*kernel;
	memset(&Gamma_seq,0,sizeof(Gamma_seq));
	if(*ngamma){
		Gamma_seq.nparam = *ngamma;
		Gamma_seq.param = gammas;
	}else{
		Gamma_seq.from = *from_gamma;
		Gamma_seq.to   = *to_gamma;
		Gamma_seq.by   = *by_gamma;
	}
// check rts param sequence
	RTS_PARAM_SEQ* gamma_seq = &Gamma_seq;
	if((gamma_seq->nparam <= 0) && (gamma_seq->param)){
		Rprintf("Error: Parameter sequence for 'gamma' has incorrectly specified length (%i)\n",0);	
		return;
	}
	if(gamma_seq->nparam == 0)
		gamma_seq->nparam = 1 + (int)ceil((gamma_seq->to - gamma_seq->from) / gamma_seq->by);
	if(!isfinite(gamma_seq->nparam)){
		Rprintf("Error: Parameter sequence for kernel parameters 'gamma' is incorrectly specified\n");
		return;
	}	
	
	memset(&Lambda_seq,0,sizeof(Lambda_seq));
	if(*nlambda){
		Lambda_seq.nparam = *nlambda;
		Lambda_seq.param = lambdas;
	}else{
		Lambda_seq.from = *from_lambda;
		Lambda_seq.to   = *to_lambda;
		Lambda_seq.by   = *by_lambda;
	}
// check rts param sequence
	RTS_PARAM_SEQ* lambda_seq = &Lambda_seq;
	if((lambda_seq->nparam <= 0) && (lambda_seq->param)){
		Rprintf("Error: Parameter sequence for penalty parameter 'lambda' has incorrectly specified length (%i)\n",0);	
		return;
	}
	if(lambda_seq->nparam == 0)
		lambda_seq->nparam = 1 + (int)ceil((lambda_seq->to - lambda_seq->from) / lambda_seq->by);
	if(!isfinite(lambda_seq->nparam) || (lambda_seq->nparam < 0)){
		Rprintf("Error: Parameter sequence for penalty parameters 'lambda' is incorrectly specified\n");
		return;
	}		
	
	
	// Rprintf("Calling cv_lambda\n");
	cv_rts(*kfold,*random,*cost,sampling_weights,lambda_seq,gamma_seq,(RTS_METHOD)*rts_method,&Kernel,&Param,&mf,*verbose,result);	
}
