#ifndef RTS_h
#define RTS_h
#include "kernels.h"
#include "loqo.h"

typedef enum {RTS_W1 = 0,RTS_SVM = 1,RTS_W2 = 2}RTS_METHOD;

typedef struct rts_model_frame_struct{
	int nrx;
	int ncx;
	double* y;
	double* X;
	double* w;
	double* tt;
}RTS_MODEL_FRAME;

typedef struct rts_control_struct{
	int verbose;
	int maxit;
	double tol;
}RTS_CONTROL;

typedef struct rts_param_struct{
	int is_intercept; // whether to compute beta[0]
	double s;
	double lambda;
	double threshold;
	// double cost;
	
// initial values, if NULL then ignored
	int neta;
	double* eta_init;  // length nrx
	int nbeta;
	double* beta_init; // length ncx + 1		

// memory	
	int aux_buffer_size;
	double* aux_buffer;	
	
// control parameters for rts and the quadratic optimizer
	RTS_CONTROL control;	
	LOQO_CONTROL loqo_control;

}RTS_PARAM;

typedef struct rts_fit_struct{
// data
	// int ncx;
	// int nrx;
	// double* X; // does not contain intercept, X <-> beta[-0]
	// double* y;
	// double* w;
	// double* tt;
	RTS_MODEL_FRAME data;

// parameters	
	RTS_PARAM param;
	KERNEL_PARAM kernel;
	
// output	
	int converged;      // 0/1 failure/success
	int convergence;    // 0 good, !=0 bad
	double* beta;       // length ncx + 1, always contains intercept 
	double* alpha_pred; // 'alpha_pred', predicted alpha
	double* linear_combination;
	double* eta;
	
}RTS_FIT;

typedef struct param_sequence{
	int nparam;     // if 0 then compute from (from,to,by) 
	double* param;  // if NULL then (from,to,by) are used 
	double from;
	double to;
	double by;
}RTS_PARAM_SEQ;

#endif

void init_control_rts(RTS_CONTROL* control);
void print_control_rts(RTS_CONTROL* control);
void init_param_rts(RTS_PARAM* param);
void print_param_rts(RTS_PARAM* param);
void print_mf_rts(RTS_MODEL_FRAME* mf,int print_X);


void rts_caltheta_weights(int* n,double* Y,double* weights,double* TT,double* score,double* cost,double* theta);
void rtsw(RTS_MODEL_FRAME*,RTS_PARAM*,KERNEL_PARAM*,RTS_FIT* arg);
void rtswt(RTS_MODEL_FRAME*,RTS_PARAM*,KERNEL_PARAM*,RTS_FIT* arg);
void rtswt_svm(RTS_MODEL_FRAME*,RTS_PARAM*,KERNEL_PARAM*,RTS_FIT* arg);

void predict_rtsw(RTS_FIT* fit,int nrx,double* x,double* result);
void predict_svmw(RTS_FIT* fit,int nrx,double* x,double* result);

void R_rtsw(
int* nrx,int* ncx,double* X,double* y,double* w,
int* kernel,double* kernel_param,
double* s,double* lambda,double* threshold,
int* is_intercept,
int* is_eta,double* eta_init,
int* is_beta,double* beta_init,

// input control
double* rts_control,
double* loqo_control,
// external memory
int* aux_buffer_size,double* aux_buffer,
// output
int* convergence,
double* beta,
double* alpha_pred,
double* linear_combination,
double* eta);


void R_rtswt(
int* nrx,int* ncx,double* X,double* y,double* w,
int* kernel,double* kernel_param,
double* s,double* lambda,double* threshold,
int* is_intercept,
int* is_eta,double* eta_init,
int* is_beta,double* beta_init,

// input control
double* rts_control,
double* loqo_control,
// external memory
int* aux_buffer_size,double* aux_buffer,
// output
int* convergence,
double* beta,
double* alpha_pred,
double* linear_combination,
double* eta);

void R_rtswt_svm(
int* nrx,int* ncx,double* X,double* y,double* w,
int* kernel,double* kernel_param,
double* s,double* lambda,double* threshold,
int* is_intercept,
int* is_eta,double* eta_init,
int* is_beta,double* beta_init,

// input control
double* rts_control,
double* loqo_control,
// external memory
int* aux_buffer_size,double* aux_buffer,
// output
int* convergence,
double* beta,
double* alpha_pred,
double* linear_combination,
double* eta);



void R_predict_rtsw2(
int* ncx,
int* nrx1,double* X1,                 // newdata_(nrx1 x ncx) on which prediction is based
int* nrx,double* X,                   // data_(nrx x ncx) used in fit 
double* y,
double* w,                            // (+/-)weights_{nrx x 1} used in fitting
double* alpha,						  // estimated alpha_{nrx x 1} from fitting
double* beta,                         // estimated coefficient beta_{ncx x 1} from fitting 
int* kernel,double* kernel_param,     // kernel and its parameter 
double* lambda,                      // tuning parameters
int* aux_buffer_size,double* aux_buffer,     // buffer needed for calculation
double* linear_combination            // prediction output of length nrx1
);

void R_predict_svml2(
int* ncx,
int* nrx1,double* X1,                 // newdata_(nrx1 x ncx) on which prediction is based
int* nrx,double* X,                   // data_(nrx x ncx) used in fit 
double* y,
double* w,                            // (+/-)weights_{nrx x 1} used in fitting
double* alpha,						  // estimated alpha_{nrx x 1} from fitting
double* beta,                         // estimated coefficient beta_{ncx x 1} from fitting 
int* kernel,double* kernel_param,     // kernel and its parameter 
double* lambda,                      // tuning parameters
int* aux_buffer_size,double* aux_buffer,     // buffer needed for calculation
double* linear_combination            // prediction output of length nrx1
);

// RTS_ARG
// RTS_FIT
// OPTIM_ARG
// KERNEL_ARG,KERNEL_PARAM
// RTS_NEW_FIT = predict(RTS_FIT,NEW_DATA)


// fit input:
// int* nrx,int* ncx,
// double* X,double* y,double* w,
// int* kernel,double* kernel_param,
// double* _s,double* _lambda,double* _threshold,
// int* is_intercept,
// int* is_eta,double* eta_init,
// int* is_beta,double* beta_init,
// int* lapack_solver_type,int* update_solver,
// int* _external_buffer_size,double* _buffer,double* _inf,
// int* _maxit,double* _tol,int* _verbose,

// fit output:
// all of the above and 
// int* convergence,
// double* beta,
// double* alpha_pred,
// double* linear_combination,
// double* eta
